import time

from rich import box
from rich.console import Console
from rich.layout import Layout
from rich.live import Live
from rich.panel import Panel
from rich.spinner import Spinner
from rich.style import Style
from rich.table import Table
from rich.text import Text

from BoatBuddy import utils
from BoatBuddy.email_manager import EmailManager
from BoatBuddy.generic_plugin import PluginStatus
from BoatBuddy.notifications_manager import NotificationsManager, EntryType
from BoatBuddy.plugin_manager import PluginManager, PluginManagerStatus
from BoatBuddy.sound_manager import SoundManager


class ConsoleManager:

    def __init__(self, options, plugin_manager: PluginManager, notifications_manager: NotificationsManager,
                 sound_manager: SoundManager, email_manager: EmailManager):
        self._options = options
        self._plugin_manager = plugin_manager
        self._notifications_manager = notifications_manager
        self._sound_manager = sound_manager
        self._email_manager = email_manager

        self._console = Console()

        with Live(self._make_layout(), refresh_per_second=4) as live:
            try:
                while True:
                    time.sleep(0.5)
                    live.update(self._make_layout())
            except KeyboardInterrupt:  # on keyboard interrupt...
                utils.get_logger().warning("Ctrl+C signal detected!")
            except Exception as e:
                utils.get_logger().error(f'An unexpected error has occurred and application will shutdown. Details {e}')
            finally:
                # Notify the plugin manager
                self._plugin_manager.finalize()
                # Notify the notifications manager
                self._notifications_manager.finalize()
                # Notify the sound manager
                self._sound_manager.finalize()
                # Notify the email manager
                self._email_manager.finalize()

    def _make_header(self) -> Layout:
        application_name = utils.get_application_name()
        application_version = utils.get_application_version()
        curr_time = time.strftime("%H:%M", time.localtime())
        status = self._plugin_manager.get_status()

        status_renderable = None
        application_info_renderable = None
        local_time_renderable = None
        if status == PluginManagerStatus.IDLE:
            status_style = Style(color='bright_white', bgcolor='default')
            status_renderable = Spinner('simpleDots', text=Text('Idle', style=status_style))
            application_info_style = Style(color='blue', bgcolor='default', bold=True)
            application_info_renderable = Text(f'{application_name} ({application_version})',
                                               style=application_info_style)
            local_time_style = Style(color='bright_yellow', bgcolor='default')
            local_time_renderable = Text(f'Local time: {curr_time}',
                                         style=local_time_style)
        elif status == PluginManagerStatus.SESSION_ACTIVE:
            status_style = Style(color='bright_white', bgcolor='red')
            status_renderable = Spinner('earth', text=Text('Session active', style=status_style))
            application_info_style = Style(color='bright_white', bgcolor='red', bold=True)
            application_info_renderable = Text(f'{application_name} ({application_version})',
                                               style=application_info_style)
            local_time_style = Style(color='bright_white', bgcolor='red')
            local_time_renderable = Text(f'Local time: {curr_time}',
                                         style=local_time_style)

        grid = Table.grid(expand=True)
        grid.add_column(justify="left")
        grid.add_column(justify="center", ratio=1, style='blue')
        grid.add_column(justify="right", style="bright_yellow")
        grid.add_row(
            status_renderable,
            application_info_renderable,
            local_time_renderable
        )
        return Layout(grid)

    def _make_summary(self) -> Layout:
        layout = Layout()
        layout.split_column(
            Layout(name="summary_header", size=5),
            Layout(name="summary_body", ratio=1),
        )

        summary_header_table = Table.grid(expand=True)
        summary_header_table.add_column()
        summary_header_table.add_column()
        summary_header_key_value_list = self._plugin_manager.get_filtered_session_clock_metrics()
        counter = 0
        while counter < len(summary_header_key_value_list):
            key = list(summary_header_key_value_list.keys())[counter]

            if counter + 1 < len(summary_header_key_value_list):
                next_key = list(summary_header_key_value_list.keys())[counter + 1]
                summary_header_table.add_row(
                    f'[bright_white]{key}: {summary_header_key_value_list[key]}[/bright_white]',
                    f'[bright_white]{next_key}: {summary_header_key_value_list[next_key]}[/bright_white]')
            else:
                summary_header_table.add_row(
                    f'[bright_white]{key}: {summary_header_key_value_list[key]}[/bright_white]', '')
            counter += 2

        layout["summary_header"].update(
            Layout(Panel(summary_header_table, title=f'{self._plugin_manager.get_session_name()}')))

        summary_body_table = Table.grid(expand=True)
        summary_body_table.add_column()
        summary_body_table.add_column()
        summary_key_value_list = self._plugin_manager.get_filtered_summary_metrics()
        counter = 0
        while counter < len(summary_key_value_list):
            key = list(summary_key_value_list.keys())[counter]

            if counter + 1 < len(summary_key_value_list):
                next_key = list(summary_key_value_list.keys())[counter + 1]
                summary_body_table.add_row(f'[bright_white]{key}: {summary_key_value_list[key]}[/bright_white]',
                                           f'[bright_white]{next_key}: {summary_key_value_list[next_key]}[/bright_white]')
            else:
                summary_body_table.add_row(f'[bright_white]{key}: {summary_key_value_list[key]}[/bright_white]', '')
            counter += 2

        layout["summary_body"].update(Layout(Panel(summary_body_table,
                                                   title=f'Session Summary')))
        return layout

    @staticmethod
    def _make_footer() -> Panel:
        footer_table = Table.grid(expand=True)
        footer_table.add_column()
        last_log_entries = utils.get_last_log_entries(3)
        for entry in last_log_entries:
            colour = 'default'
            if 'INFO' in str(entry).upper():
                colour = 'green'
            elif 'WARNING' in str(entry).upper():
                colour = 'yellow'
            elif 'ERROR' in str(entry).upper():
                colour = 'red'
            footer_table.add_row(f'[{colour}]{entry}[/{colour}]')
        return Panel(footer_table, title=f'Last 3 log entries')

    def _make_key_value_table(self, title, key_value_list, border_style) -> Panel:
        table = Table.grid(expand=True)
        table.add_column()
        colour = 'default'

        for key in key_value_list:
            if self._options.notification_console:
                colour = utils.get_colour_for_key_value_in_dictionary(self._options.metrics_colouring_scheme, key,
                                                                      key_value_list[key])
            if colour != 'default':
                table.add_row(f'[b][{colour}]{key}: ' +
                              f'{key_value_list[key]}[/{colour}][/b]')
            else:
                table.add_row(f'[bright_white]{key}: ' +
                              f'{key_value_list[key]}[/bright_white]')
            self._notifications_manager.notify(key, key_value_list[key], EntryType.METRIC)
        return Panel(table, title=title, border_style=border_style, box=box.ROUNDED)

    def _make_layout(self) -> Layout:
        layout = Layout()

        if self._options.log_module and self._options.console_show_log:
            layout.split_column(
                Layout(name="header", size=1),
                Layout(name="body", ratio=1),
                Layout(name="footer", size=5)
            )

            layout["footer"].update(self._make_footer())
        else:
            layout.split_column(
                Layout(name="header", size=1),
                Layout(name="body", ratio=1),
            )

        layout["header"].update(self._make_header())

        # Add the plugins layout for the loaded plugins
        victron_layout = None
        gps_layout = None
        nmea_layout = None
        summary_layout = None

        if self._options.victron_module and self._options.console_show_victron_plugin:
            victron_layout = Layout(name="victron")
            # Populate the victron layout
            plugin_status = self._plugin_manager.get_victron_plugin_status()
            formatted_plugin_status_str = self._get_formatted_plugin_status_str(plugin_status)
            border_style = self._get_border_style_from_status(plugin_status)
            victron_layout.update(self._make_key_value_table('Victron ESS ' + formatted_plugin_status_str,
                                                             self._plugin_manager.get_filtered_victron_metrics(),
                                                             border_style))

        if self._options.gps_module and self._options.console_show_gps_plugin:
            gps_layout = Layout(name="gps")
            # Populate the NMEA layout
            plugin_status = self._plugin_manager.get_gps_plugin_status()
            formatted_plugin_status_str = self._get_formatted_plugin_status_str(plugin_status)
            border_style = self._get_border_style_from_status(plugin_status)
            gps_layout.update(self._make_key_value_table('GPS Module ' + formatted_plugin_status_str,
                                                         self._plugin_manager.get_filtered_gps_metrics(),
                                                         border_style))

        if self._options.nmea_module and self._options.console_show_nmea_plugin:
            nmea_layout = Layout(name="nmea")
            # Populate the NMEA layout
            plugin_status = self._plugin_manager.get_nmea_plugin_status()
            formatted_plugin_status_str = self._get_formatted_plugin_status_str(plugin_status)
            border_style = self._get_border_style_from_status(plugin_status)
            nmea_layout.update(self._make_key_value_table('NMEA0183 Network ' + formatted_plugin_status_str,
                                                          self._plugin_manager.get_filtered_nmea_metrics(),
                                                          border_style))

        if self._plugin_manager.get_status() == PluginManagerStatus.SESSION_ACTIVE:
            summary_layout = Layout(name="summary", ratio=2)
            summary_layout.update(self._make_summary())

        layouts_list = []
        if victron_layout:
            layouts_list.append(victron_layout)

        if nmea_layout:
            layouts_list.append(nmea_layout)

        if gps_layout:
            layouts_list.append(gps_layout)

        if summary_layout:
            layouts_list.append(summary_layout)

        layout["body"].split_row(*layouts_list)

        return layout

    @staticmethod
    def _get_formatted_plugin_status_str(plugin_status: PluginStatus):
        plugin_status_str = ''
        if plugin_status == PluginStatus.DOWN:
            plugin_status_str = '[red](Down)[/red]'
        elif plugin_status == PluginStatus.STARTING:
            plugin_status_str = '[bright_yellow](Starting)[/bright_yellow]'
        elif plugin_status == PluginStatus.RUNNING:
            plugin_status_str = '[green](Running)[/green]'
        return plugin_status_str

    @staticmethod
    def _get_border_style_from_status(plugin_status: PluginStatus):
        border_style = 'default'
        if plugin_status == PluginStatus.DOWN:
            border_style = 'red'
        elif plugin_status == PluginStatus.STARTING:
            border_style = 'bright_yellow'
        elif plugin_status == PluginStatus.RUNNING:
            border_style = 'green'

        return border_style



import os
from enum import Enum
from threading import Thread, Event, Lock

from playsound import playsound

from BoatBuddy import utils


class SoundType(Enum):
    WARNING = 1
    ALARM = 2
    APPLICATION_STARTED = 3
    SESSION_STARTED = 4
    SESSION_ENDED = 5


class SoundManager:

    def __init__(self, options):
        self._options = options
        self._sound_queue = []
        self._sound_thread = None
        self._exit_signal = Event()
        self._sound_thread = Thread(target=self._main_loop)

        if self._options.sound_module:
            self._mutex = Lock()
            self._sound_thread.start()
            utils.get_logger().info('Sound module successfully started!')

    def play_sound_async(self, sound_type: SoundType):
        if not self._options.sound_module:
            return
        filename = None
        if sound_type == SoundType.APPLICATION_STARTED:
            filename = '/resources/application_started.mp3'
        elif sound_type == SoundType.ALARM:
            filename = '/resources/alarm.mp3'
        elif sound_type == SoundType.WARNING:
            filename = '/resources/warning.mp3'
        elif sound_type == SoundType.SESSION_STARTED:
            filename = '/resources/session_started.mp3'
        elif sound_type == SoundType.SESSION_ENDED:
            filename = '/resources/session_ended.wav'

        if filename:
            self._mutex.acquire()
            if len(self._sound_queue) == 0:
                self._sound_queue.append(filename)
            elif self._sound_queue[len(self._sound_queue) - 1] != filename:
                self._sound_queue.append(filename)
            self._mutex.release()

    def finalize(self):
        if not self._options.sound_module:
            return

        self._exit_signal.set()
        if self._sound_thread:
            self._sound_thread.join()

        utils.get_logger().info('Sound manager instance is ready to be destroyed')

    def _main_loop(self):
        while not self._exit_signal.is_set():
            self._mutex.acquire()
            if len(self._sound_queue):
                self._play_sound(self._sound_queue.pop(0))
            self._mutex.release()

    @staticmethod
    def _play_sound(filename):
        full_path = os.path.dirname(os.path.abspath(__file__)) + filename
        utils.get_logger().debug(f'Playing a sound with filename: {full_path}')
        playsound(full_path)



from enum import Enum


class PluginStatus(Enum):
    STARTING = 1
    RUNNING = 2
    DOWN = 3


class GenericPlugin:
    _options = None

    def __init__(self, options):
        self._options = options
        self._log_entries = []

    def get_metadata_headers(self):
        return []

    # Collect all current data in an object in memory (add that object to a list instance if needed)
    def take_snapshot(self, store_entry):
        raise NotImplementedError("Method needs to be implemented")

    def get_metadata_values(self):
        return []

    def get_summary_headers(self):
        return []

    def get_summary_values(self):
        return []

    def clear_entries(self):
        self._log_entries = []

    # Close active sessions (if any), this method is called when a KeyboardInterrupt signal is raised
    def finalize(self):
        raise NotImplementedError("Method needs to be implemented")

    def register_for_events(self, events):
        pass

    def get_status(self) -> PluginStatus:
        raise NotImplementedError("Method needs to be implemented")



import threading

import numpy as np
from events import Events
from pyModbusTCP.client import ModbusClient

from BoatBuddy import globals, utils
from BoatBuddy.generic_plugin import GenericPlugin, PluginStatus


class VictronPluginEvents(Events):
    __events__ = ('on_connect', 'on_disconnect',)


class VictronEntry:
    def __init__(self, input_source_string, grid_power, generator_power, ac_input_voltage, ac_input_current,
                 ac_input_frequency, ve_bus_state_string, ac_consumption, battery_voltage,
                 battery_current, battery_power, battery_soc, battery_state_string,
                 pv_power, pv_current, starter_battery_voltage, tank1_level, tank1_type, tank2_level, tank2_type):
        self._input_source_string = input_source_string
        self._grid_power = grid_power
        self._generator_power = generator_power
        self._ac_input_voltage = ac_input_voltage
        self._ac_input_current = ac_input_current
        self._ac_input_frequency = ac_input_frequency
        self._ve_bus_state_string = ve_bus_state_string
        self._ac_consumption = ac_consumption
        self._battery_voltage = battery_voltage
        self._battery_current = battery_current
        self._battery_power = battery_power
        self._battery_soc = battery_soc
        self._battery_state_string = battery_state_string
        self._pv_power = pv_power
        self._pv_current = pv_current
        self._starter_battery_voltage = starter_battery_voltage
        self._tank1_level = tank1_level
        self._tank1_type = tank1_type
        self._tank2_level = tank2_level
        self._tank2_type = tank2_type

    def __str__(self):
        return utils.get_comma_separated_string(self.get_values())

    def get_values(self):
        return [f'{self._input_source_string}', f'{self._grid_power}', f'{self._generator_power}',
                f'{self._ac_input_voltage}', f'{self._ac_input_current}', f'{self._ac_input_frequency}',
                f'{self._ve_bus_state_string}', f'{self._ac_consumption}', f'{self._battery_voltage}',
                f'{self._battery_current}', f'{self._battery_power}', f'{self._battery_soc}',
                f'{self._battery_state_string}',
                f'{self._pv_power}', f'{self._pv_current}', f'{self._starter_battery_voltage}',
                f'{self._tank1_level}', f'{self._tank1_type}', f'{self._tank2_level}', f'{self._tank2_type}']

    def get_battery_voltage(self):
        return self._battery_voltage

    def get_battery_current(self):
        return self._battery_current

    def get_battery_power(self):
        return self._battery_power

    def get_pv_power(self):
        return self._pv_power

    def get_pv_current(self):
        return self._pv_current

    def get_starter_battery_voltage(self):
        return self._starter_battery_voltage

    def get_ac_consumption_power(self):
        return self._ac_consumption

    def get_tank1_level(self):
        return self._tank1_level

    def get_tank1_type(self):
        return self._tank1_type

    def get_tank2_level(self):
        return self._tank2_level

    def get_tank2_type(self):
        return self._tank2_type


class VictronPlugin(GenericPlugin):
    _events = None

    def __init__(self, options):
        # invoking the __init__ of the parent class
        GenericPlugin.__init__(self, options)

        # Instance metrics
        self._grid_power = ''
        self._generator_power = ''
        self._input_source_string = ''
        self._ac_consumption = ''
        self._battery_state_string = ''
        self._battery_voltage = ''
        self._battery_current = ''
        self._battery_power = ''
        self._battery_soc = ''
        self._pv_power = ''
        self._pv_current = ''
        self._starter_battery_voltage = ''
        self._ac_input_voltage = ''
        self._ac_input_current = ''
        self._ac_input_frequency = ''
        self._ve_bus_state_string = ''
        self._tank1_level = ''
        self._tank1_type_string = ''
        self._tank2_level = ''
        self._tank2_type_string = ''

        # Other instance variables
        self._plugin_status = PluginStatus.STARTING
        self._exit_signal = threading.Event()
        self._timer = threading.Timer(globals.VICTRON_TIMER_INTERVAL, self.main_loop)
        self._timer.start()
        utils.get_logger().info('Victron module successfully started!')

    def reset_instance_metrics(self):
        self._grid_power = ''
        self._generator_power = ''
        self._input_source_string = ''
        self._ac_consumption = ''
        self._battery_state_string = ''
        self._battery_voltage = ''
        self._battery_current = ''
        self._battery_power = ''
        self._battery_soc = ''
        self._pv_power = ''
        self._pv_current = ''
        self._starter_battery_voltage = ''
        self._ac_input_voltage = ''
        self._ac_input_current = ''
        self._ac_input_frequency = ''
        self._ve_bus_state_string = ''
        self._tank1_level = ''
        self._tank1_type_string = ''
        self._tank2_level = ''
        self._tank2_type_string = ''

    def main_loop(self):
        if self._exit_signal.is_set():
            self._plugin_status = PluginStatus.DOWN
            utils.get_logger().info('Victron plugin instance is ready to be destroyed')
            return

        server_ip = f'{self._options.victron_server_ip}'
        server_port = self._options.victron_tcp_port

        try:
            # TCP auto connect on modbus request, close after it
            c = ModbusClient(host=server_ip, port=server_port, unit_id=100, auto_open=True, auto_close=True)

            # Try to make a test inquiry to the system before proceeding any further
            if c.read_holding_registers(820, 1) is None:
                raise ValueError('Modbus TCP server is unreachable')

            if self._plugin_status != PluginStatus.RUNNING:
                utils.get_logger().info(f'Connection to Victron system on {server_ip} is established')

                self._plugin_status = PluginStatus.RUNNING

                self.reset_instance_metrics()

                if self._events:
                    self._events.on_connect()

            try:
                self._grid_power = utils.try_parse_int(c.read_holding_registers(820, 1)[0])
                self._generator_power = utils.try_parse_int(np.int16(c.read_holding_registers(823, 1)[0]))

                self._input_source_string = ''
                input_source = utils.try_parse_int(c.read_holding_registers(826, 1)[0])
                if input_source == 0:
                    self._input_source_string = 'Unknown'
                elif input_source == 1:
                    self._input_source_string = 'Grid'
                elif input_source == 2:
                    self._input_source_string = 'Generator'
                elif input_source == 3:
                    self._input_source_string = 'Shore Power'
                elif input_source == 240:
                    self._input_source_string = 'Not Connected'

                self._ac_consumption = utils.try_parse_int(c.read_holding_registers(817, 1)[0])

                self._battery_state_string = ''
                battery_state = utils.try_parse_int(c.read_holding_registers(844, 1)[0])
                if battery_state == 0:
                    self._battery_state_string = 'idle'
                elif battery_state == 1:
                    self._battery_state_string = 'charging'
                elif battery_state == 2:
                    self._battery_state_string = 'discharging'

                self._battery_voltage = utils.try_parse_int(c.read_holding_registers(840, 1)[0]) / 10
                self._battery_current = utils.try_parse_int(np.int16(c.read_holding_registers(841, 1)[0])) / 10
                self._battery_power = utils.try_parse_int(np.int16(c.read_holding_registers(842, 1)[0]))
                self._battery_soc = utils.try_parse_int(c.read_holding_registers(843, 1)[0])
                self._pv_power = utils.try_parse_int(c.read_holding_registers(850, 1)[0])
                self._pv_current = utils.try_parse_int(np.int16(c.read_holding_registers(851, 1)[0])) / 10

                # Get starter battery voltage
                c.unit_id = 223
                self._starter_battery_voltage = utils.try_parse_int(c.read_holding_registers(260, 1)[0]) / 100

                # Get VE.Bus metrics
                c.unit_id = 227
                self._ac_input_voltage = utils.try_parse_int(c.read_holding_registers(3, 1)[0]) / 10
                self._ac_input_current = utils.try_parse_int(np.int16(c.read_holding_registers(6, 1)[0])) / 10
                self._ac_input_frequency = utils.try_parse_int(c.read_holding_registers(9, 1)[0]) / 100

                self._ve_bus_state_string = ''
                ve_bus_state = utils.try_parse_int(c.read_holding_registers(31, 1)[0])
                if ve_bus_state == 0:
                    self._ve_bus_state_string = 'Off'
                elif ve_bus_state == 1:
                    self._ve_bus_state_string = 'Low Power'
                elif ve_bus_state == 2:
                    self._ve_bus_state_string = 'Fault'
                elif ve_bus_state == 3:
                    self._ve_bus_state_string = 'Bulk'
                elif ve_bus_state == 4:
                    self._ve_bus_state_string = 'Absorption'
                elif ve_bus_state == 5:
                    self._ve_bus_state_string = 'Float'
                elif ve_bus_state == 6:
                    self._ve_bus_state_string = 'Storage'
                elif ve_bus_state == 7:
                    self._ve_bus_state_string = 'Equalize'
                elif ve_bus_state == 8:
                    self._ve_bus_state_string = 'Passthru'
                elif ve_bus_state == 9:
                    self._ve_bus_state_string = 'Inverting'
                elif ve_bus_state == 10:
                    self._ve_bus_state_string = 'Power assist'
                elif ve_bus_state == 11:
                    self._ve_bus_state_string = 'Power supply'
                elif ve_bus_state == 252:
                    self._ve_bus_state_string = 'External control'

                c.unit_id = 20
                self._tank1_level = utils.try_parse_int(utils.try_parse_int(c.read_holding_registers(3004, 1)[0]) / 10)
                tank1_type = utils.try_parse_int(c.read_holding_registers(3003, 1)[0])
                self._tank1_type_string = ''
                # 0=Fuel;1=Fresh water;2=Waste water;3=Live well;4=Oil;5=Black water (sewage);
                # 6=Gasoline;7=Diesel;8=LPG;9=LNG;10=Hydraulic oil;11=Raw water
                if tank1_type == 0:
                    self._tank1_type_string = 'Fuel'
                elif tank1_type == 1:
                    self._tank1_type_string = 'Fresh water'
                elif tank1_type == 2:
                    self._tank1_type_string = 'Waste water'
                elif tank1_type == 3:
                    self._tank1_type_string = 'Live well'
                elif tank1_type == 4:
                    self._tank1_type_string = 'Oil'
                elif tank1_type == 5:
                    self._tank1_type_string = 'Black water (sewage)'
                elif tank1_type == 6:
                    self._tank1_type_string = 'Gasoline'
                elif tank1_type == 7:
                    self._tank1_type_string = 'Diesel'
                elif tank1_type == 8:
                    self._tank1_type_string = 'LPG'
                elif tank1_type == 9:
                    self._tank1_type_string = 'LNG'
                elif tank1_type == 10:
                    self._tank1_type_string = 'Hydraulic oil'
                elif tank1_type == 11:
                    self._tank1_type_string = 'Raw water'

                c.unit_id = 21
                self._tank2_level = utils.try_parse_int(utils.try_parse_int(c.read_holding_registers(3004, 1)[0]) / 10)
                tank2_type = utils.try_parse_int(c.read_holding_registers(3003, 1)[0])
                self._tank2_type_string = ''
                if tank2_type == 0:
                    self._tank2_type_string = 'Fuel'
                elif tank2_type == 1:
                    self._tank2_type_string = 'Fresh water'
                elif tank2_type == 2:
                    self._tank2_type_string = 'Waste water'
                elif tank2_type == 3:
                    self._tank2_type_string = 'Live well'
                elif tank2_type == 4:
                    self._tank2_type_string = 'Oil'
                elif tank2_type == 5:
                    self._tank2_type_string = 'Black water (sewage)'
                elif tank2_type == 6:
                    self._tank2_type_string = 'Gasoline'
                elif tank2_type == 7:
                    self._tank2_type_string = 'Diesel'
                elif tank2_type == 8:
                    self._tank2_type_string = 'LPG'
                elif tank2_type == 9:
                    self._tank2_type_string = 'LNG'
                elif tank2_type == 10:
                    self._tank2_type_string = 'Hydraulic oil'
                elif tank2_type == 11:
                    self._tank2_type_string = 'Raw water'
            except Exception as e:
                self._handle_connection_exception(e)
        except Exception as e:
            self._handle_connection_exception(e)

        # Reset the timer
        self._timer = threading.Timer(globals.VICTRON_TIMER_INTERVAL, self.main_loop)
        self._timer.start()

    def _handle_connection_exception(self, message):
        if self._plugin_status != PluginStatus.DOWN:
            utils.get_logger().info(
                f'Problem with Victron system on {self._options.victron_server_ip}. Details: {message}')

            self._plugin_status = PluginStatus.DOWN

            # If anyone is listening to events then notify of a disconnection
            if self._events:
                self._events.on_disconnect()

    def get_metadata_headers(self):
        return globals.VICTRON_PLUGIN_METADATA_HEADERS.copy()

    def take_snapshot(self, store_entry):
        entry = VictronEntry(self._input_source_string, self._grid_power, self._generator_power,
                             self._ac_input_voltage, self._ac_input_current, self._ac_input_frequency,
                             self._ve_bus_state_string, self._ac_consumption, self._battery_voltage,
                             self._battery_current, self._battery_power, self._battery_soc,
                             self._battery_state_string, self._pv_power, self._pv_current,
                             self._starter_battery_voltage, self._tank1_level, self._tank1_type_string,
                             self._tank2_level, self._tank2_type_string)

        if store_entry:
            utils.get_logger().debug(f'Adding new Victron entry')
            utils.get_logger().debug(f'Active Input source: {self._input_source_string} ' +
                                     f'Grid Power: {self._grid_power} W ' +
                                     f'Generator Power: {self._generator_power} W ' +
                                     f'AC Consumption: {self._ac_consumption} W')
            utils.get_logger().debug(f'AC input 1 {self._ac_input_voltage} V {self._ac_input_current} A ' +
                                     f'{self._ac_input_frequency} Hz ' +
                                     f'State: {self._ve_bus_state_string}')
            utils.get_logger().debug(
                f'Housing battery stats {self._battery_voltage} V  {self._battery_current} A {self._battery_power} W ' +
                f'{self._battery_soc} % {self._battery_state_string}')
            utils.get_logger().debug(f'PV {self._pv_power} W {self._pv_current} A')
            utils.get_logger().debug(f'Starter battery voltage: {self._starter_battery_voltage} V')
            utils.get_logger().debug(f'Tank 1 Level: {self._tank1_level} Type: {self._tank1_type_string}')
            utils.get_logger().debug(f'Tank 2 Level: {self._tank2_level} Type: {self._tank2_type_string}')

            self._log_entries.append(entry)

        return entry

    def get_metadata_values(self):
        if len(self._log_entries) > 0:
            return self._log_entries[len(self._log_entries) - 1].get_values()
        else:
            return []

    def finalize(self):
        self._exit_signal.set()
        if self._timer:
            self._timer.cancel()
        utils.get_logger().info("Victron plugin worker thread notified...")

    def get_summary_headers(self):
        return globals.VICTRON_PLUGINS_SUMMARY_HEADERS.copy()

    def get_summary_values(self):
        log_summary_list = []

        if len(self._log_entries) > 0:
            # Calculate extremes and averages
            housing_battery_max_voltage = 0
            housing_battery_min_voltage = 0
            sum_housing_battery_voltage = 0
            housing_battery_max_current = 0
            sum_housing_battery_current = 0
            housing_battery_max_power = 0
            sum_housing_battery_power = 0
            pv_max_power = 0
            sum_pv_power = 0
            pv_max_current = 0
            sum_pv_current = 0
            starter_battery_max_voltage = 0
            starter_battery_min_voltage = 0
            sum_starter_battery_voltage = 0
            ac_consumption_max_power = 0
            sum_ac_consumption_power = 0
            tank1_min_level = 0
            tank1_max_level = 0
            sum_tank1_level = 0
            tank2_min_level = 0
            tank2_max_level = 0
            sum_tank2_level = 0
            count = len(self._log_entries)
            for entry in self._log_entries:
                # Sum up all values that will be averaged later
                sum_housing_battery_voltage += utils.try_parse_float(entry.get_battery_voltage())
                sum_housing_battery_current += utils.try_parse_float(entry.get_battery_current())
                sum_housing_battery_power += utils.try_parse_float(entry.get_battery_power())
                sum_pv_power += utils.try_parse_float(entry.get_pv_power())
                sum_pv_current += utils.try_parse_float(entry.get_pv_current())
                sum_starter_battery_voltage += utils.try_parse_float(entry.get_starter_battery_voltage())
                sum_ac_consumption_power += utils.try_parse_float(entry.get_ac_consumption_power())
                sum_tank1_level += utils.try_parse_int(entry.get_tank1_level())
                sum_tank2_level += utils.try_parse_int(entry.get_tank2_level())

                # Collect extremes
                housing_battery_max_voltage = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_battery_voltage()), housing_battery_max_voltage)
                if housing_battery_min_voltage == 0:
                    housing_battery_min_voltage = utils.try_parse_float(entry.get_battery_voltage())
                else:
                    housing_battery_min_voltage = utils.get_smallest_number(
                        utils.try_parse_float(entry.get_battery_voltage()), housing_battery_min_voltage)
                housing_battery_max_current = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_battery_current()), housing_battery_max_current)
                housing_battery_max_power = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_battery_power()), housing_battery_max_power)
                pv_max_power = utils.get_biggest_number(utils.try_parse_float(entry.get_pv_power()), pv_max_power)
                pv_max_current = utils.get_biggest_number(utils.try_parse_float(entry.get_pv_current()), pv_max_current)
                starter_battery_max_voltage = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_starter_battery_voltage()), starter_battery_max_voltage)
                if starter_battery_min_voltage == 0:
                    starter_battery_min_voltage = utils.try_parse_float(entry.get_starter_battery_voltage())
                else:
                    starter_battery_min_voltage = utils.get_smallest_number(
                        utils.try_parse_float(entry.get_starter_battery_voltage()), starter_battery_min_voltage)
                ac_consumption_max_power = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_ac_consumption_power()), ac_consumption_max_power)
                if tank1_min_level == 0:
                    tank1_min_level = utils.try_parse_float(entry.get_tank1_level())
                else:
                    tank1_min_level = utils.get_smallest_number(
                        utils.try_parse_float(entry.get_tank1_level()), tank1_min_level)
                tank1_max_level = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_tank1_level()), tank1_max_level)

                if tank2_min_level == 0:
                    tank2_min_level = utils.try_parse_float(entry.get_tank2_level())
                else:
                    tank2_min_level = utils.get_smallest_number(
                        utils.try_parse_float(entry.get_tank2_level()), tank2_min_level)
                tank2_max_level = utils.get_biggest_number(
                    utils.try_parse_float(entry.get_tank2_level()), tank2_max_level)

            log_summary_list.append(housing_battery_max_voltage)
            log_summary_list.append(housing_battery_min_voltage)
            log_summary_list.append(round(sum_housing_battery_voltage / count, 2))
            log_summary_list.append(housing_battery_max_current)
            log_summary_list.append(round(sum_housing_battery_current / count, 2))
            log_summary_list.append(housing_battery_max_power)
            log_summary_list.append(round(sum_housing_battery_power / count))
            log_summary_list.append(pv_max_power)
            log_summary_list.append(round(sum_pv_power / count))
            log_summary_list.append(pv_max_current)
            log_summary_list.append(round(sum_pv_current / count, 1))
            log_summary_list.append(starter_battery_max_voltage)
            log_summary_list.append(starter_battery_min_voltage)
            log_summary_list.append(round(sum_starter_battery_voltage / count, 2))
            log_summary_list.append(ac_consumption_max_power)
            log_summary_list.append(round(sum_ac_consumption_power / count))
            log_summary_list.append(tank1_max_level)
            log_summary_list.append(tank1_min_level)
            log_summary_list.append(round(sum_tank1_level / count))
            log_summary_list.append(tank2_max_level)
            log_summary_list.append(tank2_min_level)
            log_summary_list.append(round(sum_tank2_level / count))
        else:
            log_summary_list = ['N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A',
                                'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A', 'N/A',
                                'N/A', 'N/A']

        return log_summary_list

    def get_status(self) -> PluginStatus:
        return self._plugin_status

    def register_for_events(self, events):
        self._events = events



import threading
import time
from enum import Enum

from BoatBuddy import globals, utils
from BoatBuddy.email_manager import EmailManager
from BoatBuddy.sound_manager import SoundManager, SoundType


class EntryType(Enum):
    METRIC = 'metric'
    MODULE = 'module'


class NotificationType(Enum):
    SOUND = 'sound'
    EMAIL = 'email'


class NotificationEntry:
    def __init__(self, timestamp, key, value, entry_type: EntryType, notification_types: [], severity, frequency,
                 cool_off_interval, configuration_range=None, interval=None):
        self._timestamp = timestamp
        self._key = key
        self._value = value
        self._entry_type = entry_type
        self._notification_types = notification_types
        self._severity = severity
        self._frequency = frequency
        self._cool_off_interval = cool_off_interval
        self._configuration_range = configuration_range
        self._interval = interval

    def get_timestamp(self):
        return self._timestamp

    def get_key(self):
        return self._key

    def get_value(self):
        return self._value

    def get_entry_type(self):
        return self._entry_type

    def get_severity(self):
        return self._severity

    def get_notification_types(self):
        return self._notification_types

    def get_frequency(self):
        return self._frequency

    def get_cool_off_interval(self):
        return self._cool_off_interval

    def get_configuration_range(self):
        return self._configuration_range

    def get_interval(self):
        return self._interval


class NotificationsManager:

    def __init__(self, options, sound_manager: SoundManager, email_manager: EmailManager):
        self._options = options
        self._sound_manager = sound_manager
        self._email_manager = email_manager
        self._notifications_queue = {}
        self._exit_signal = threading.Event()
        self._notifications_thread = threading.Thread(target=self._main_loop)
        self._mutex = threading.Lock()
        if self._options.notifications_module:
            self._notifications_thread.start()
            utils.get_logger().info('Notifications module successfully started!')

    def notify(self, key, value, entry_type):
        if not self._options.notifications_module:
            return

        if entry_type == EntryType.METRIC:
            notifications_rules = self._options.metrics_notifications_rules.copy()
        elif entry_type == EntryType.MODULE:
            notifications_rules = self._options.modules_notifications_rules.copy()
        else:
            return

        # First check if the provided key has a notification configuration
        if key not in notifications_rules:
            return

        # If an empty value is provided then return
        if value is None or value == '' or str(value).upper() == 'N/A':
            return

        # Next, check if the value is falls within a range where a notification should occur
        notification_configuration = notifications_rules[key]
        for severity in notification_configuration:
            if entry_type == EntryType.METRIC:
                configuration_range = notification_configuration[severity]['range']
                if configuration_range[1] >= utils.try_parse_float(value) > configuration_range[0]:
                    notification_interval = None
                    cool_off_interval = utils.try_parse_int(self._options.notification_cool_off_interval)
                    if notification_configuration[severity]['frequency'] == 'interval':
                        notification_interval = utils.try_parse_int(notification_configuration[severity]['interval'])
                    if 'cool-off-interval' in notification_configuration[severity]:
                        cool_off_interval = utils.try_parse_int(
                            notification_configuration[severity]['cool-off-interval'])
                    self._schedule_notification(key, value, entry_type,
                                                notification_configuration[severity]['notifications'],
                                                severity, notification_configuration[severity]['frequency'],
                                                cool_off_interval, configuration_range, notification_interval)
                    return
            elif entry_type == EntryType.MODULE:
                status = notification_configuration[severity]['status']
                if value == status:
                    notification_interval = None
                    cool_off_interval = utils.try_parse_int(self._options.notification_cool_off_interval)
                    if notification_configuration[severity]['frequency'] == 'interval':
                        notification_interval = utils.try_parse_int(notification_configuration[severity]['interval'])
                    if 'cool-off-interval' in notification_configuration[severity]:
                        cool_off_interval = utils.try_parse_int(
                            notification_configuration[severity]['cool-off-interval'])
                    self._schedule_notification(key, value, entry_type,
                                                notification_configuration[severity]['notifications'],
                                                severity, notification_configuration[severity]['frequency'],
                                                cool_off_interval, None, notification_interval)
                    return

        # If this point in the code is reached then notifications for this entry (if any) should be cleared
        self._mutex.acquire()
        self._delayed_clear_notification_entry(key)
        self._mutex.release()

    def _schedule_notification(self, key, value, entry_type, notification_types, severity, frequency, cool_off_interval,
                               configuration_range=None, interval=None):
        self._mutex.acquire()
        if key not in self._notifications_queue:
            # this is a new notification entry
            self._delayed_add_notification_entry(time.time(), key, value, entry_type, notification_types, severity,
                                                 frequency, cool_off_interval, configuration_range, interval)
        elif key in self._notifications_queue and self._notifications_queue[key]['to_clear'] and \
                (self._notifications_queue[key]['instance'].get_entry_type() == EntryType.METRIC and
                 self._notifications_queue[key]['instance'].get_configuration_range() == configuration_range or
                 self._notifications_queue[key]['instance'].get_entry_type() == EntryType.MODULE and
                 self._notifications_queue[key]['instance'].get_value() == value):
            # If the item is in the queue and marked for clearance
            # Reset the item clearance status
            notification_entry = self._notifications_queue[key]['instance']
            self._notifications_queue[key]['notify_of_clearance'] = True
            self._notifications_queue[key]['to_clear'] = False
            self._notifications_queue[key]['clear_timestamp'] = None

            utils.get_logger().info(f'Notification to clear canceled for '
                                    f' {notification_entry.get_entry_type().value} \'{key}\'.')
        elif key in self._notifications_queue and \
                self._notifications_queue[key]['instance'].get_entry_type() == EntryType.METRIC and \
                self._notifications_queue[key]['instance'].get_configuration_range() != configuration_range or \
                self._notifications_queue[key]['instance'].get_entry_type() == EntryType.MODULE and \
                self._notifications_queue[key]['instance'].get_value() != value:
            # If there is already an entry in the queue with the same key
            # and if the range provided is different as what is stored in memory
            # Or if the new entry is for a module and has a different value than the old one then
            # this notification is different and needs to be treated as new notification
            # thus we need clear the old notification entry
            self._delayed_clear_notification_entry(key)
        self._mutex.release()

    def _process_notification(self, key, value, entry_type, notification_types, severity, frequency, cool_off_interval,
                              configuration_range, interval):
        if entry_type == EntryType.MODULE:
            utils.get_logger().info(f'Processing notification for module \'{key}\'')
        elif entry_type == EntryType.METRIC:
            utils.get_logger().info(f'Processing notification for metric with key \'{key}\'')

        if NotificationType.SOUND.value in notification_types:
            self._process_sound_notification(severity)

        if NotificationType.EMAIL.value in notification_types:
            self._process_email_notification(key, value, entry_type, severity, frequency, cool_off_interval,
                                             configuration_range, interval)

        # Update the last_processed field value with the current time
        self._notifications_queue[key]['last_processed'] = time.time()

    def _process_clear_notification(self, key, notification_types, severity):
        if NotificationType.EMAIL.value in notification_types:
            self._process_clear_email_notification(key, severity)

    def _process_sound_notification(self, severity):
        if severity == 'alarm':
            self._sound_manager.play_sound_async(SoundType.ALARM)
        elif severity == 'warning':
            self._sound_manager.play_sound_async(SoundType.WARNING)

    def _process_email_notification(self, key, value, entry_type, severity, frequency, cool_off_interval,
                                    configuration_range, interval):
        try:
            configuration_range_str = 'N/A'
            interval_str = 'N/A'

            if configuration_range:
                configuration_range_str = str(configuration_range)

            if interval:
                interval_str = str(interval)

            body = 'N/A'
            subject = 'N/A'
            local_time = time.strftime('%H:%M:%S')
            local_date = time.strftime('%Y-%m-%d')
            if entry_type == EntryType.MODULE:
                body = f'Notification triggered for the \'{key}\' module @ {local_time} on {local_date}:\r\n' \
                       f'Status: {value}\r\nSeverity: {severity}' + \
                       f'\r\nFrequency: {frequency}\r\nConfiguration Range: ' \
                       f'{configuration_range_str}\r\nInterval: {interval_str} seconds\r\n ' \
                       f'Cool Off Interval: {cool_off_interval} seconds\r\n\r\n' \
                       f'--\r\n{globals.APPLICATION_NAME} ({globals.APPLICATION_VERSION})'
                subject = f'{globals.APPLICATION_NAME} - ({str(severity).upper()}) ' \
                          f'notification for \'{key}\' module'
            elif entry_type == EntryType.METRIC:
                body = f'Notification triggered for metric with key \'{key}\' @ {local_time} on {local_date}:\r\n' \
                       f'Value: {value}\r\nSeverity: {severity}' + \
                       f'\r\nFrequency: {frequency}\r\nConfiguration Range: ' \
                       f'{configuration_range_str}\r\nInterval: {interval_str} seconds\r\n ' \
                       f'Cool Off Interval: {cool_off_interval} seconds\r\n\r\n' \
                       f'--\r\n{globals.APPLICATION_NAME} ({globals.APPLICATION_VERSION})'
                subject = f'{globals.APPLICATION_NAME} - ({str(severity).upper()}) ' \
                          f'notification for metric \'{key}\''
            self._email_manager.send_email(subject, body)
            if entry_type == EntryType.MODULE:
                utils.get_logger().info(f'Email notification triggered '
                                        f'for the {key} module. Status: {value} '
                                        f'Severity: {severity} Frequency: {frequency} '
                                        f'Configuration Range: {configuration_range_str} Interval: {interval_str} '
                                        f'Cool Off Interval: {cool_off_interval}. An email will be sent out shortly!')
            elif entry_type == EntryType.METRIC:
                utils.get_logger().info(f'Email notification triggered '
                                        f'for the following {entry_type.value}. Key: {key} Value: {value} '
                                        f'Severity: {severity} Frequency: {frequency} '
                                        f'Configuration Range: {configuration_range_str} Interval: {interval_str} '
                                        f'Cool Off Interval: {cool_off_interval}. An email will be sent out shortly!')
        except Exception as e:
            utils.get_logger().error(f'Error while triggering email notification for {entry_type.value} '
                                     f'\'{key}\'. Details: {e}')

    def _process_clear_email_notification(self, key, severity):
        notification_entry = self._notifications_queue[key]['instance']
        try:
            local_time = time.strftime('%H:%M:%S')
            local_date = time.strftime('%Y-%m-%d')
            body = f'Notification cleared for ' \
                   f'{notification_entry.get_entry_type().value} \'{key}\' @ {local_time} on {local_date}\r\n\r\n' \
                   f'--\r\n{globals.APPLICATION_NAME} ({globals.APPLICATION_VERSION})'
            subject = f'{globals.APPLICATION_NAME} - ({str(severity).upper()}) cleared ' \
                      f'for {notification_entry.get_entry_type().value} \'{key}\''
            self._email_manager.send_email(subject, body)
            utils.get_logger().info(f'Notification cleared for '
                                    f' {notification_entry.get_entry_type().value} \'{key}\'. '
                                    f'An email will be sent out shortly!')
        except Exception as e:
            utils.get_logger().error(f'Error while clearing email notification '
                                     f'for {notification_entry.get_entry_type().value} \'{key}\'. Details: {e}')

    def _delayed_add_notification_entry(self, timestamp, key, value, entry_type, notification_types, severity,
                                        frequency, cool_off_interval, configuration_range, interval):

        new_notification_entry = NotificationEntry(timestamp, key, value, entry_type, notification_types, severity,
                                                   frequency, cool_off_interval, configuration_range, interval)

        self._notifications_queue[key] = {'instance': new_notification_entry, 'last_processed': None,
                                          'to_clear': False, 'clear_timestamp': None, 'notify_of_clearance': True}

        if entry_type == EntryType.METRIC:
            utils.get_logger().info(f'New notification added for metric with key \'{key}\', value \'{value}\', ' +
                                    f'severity \'{severity}\'')
        elif entry_type == EntryType.MODULE:
            utils.get_logger().info(f'New notification added for module \'{key}\', status \'{value}\', ' +
                                    f'severity \'{severity}\'')

    def _delayed_clear_notification_entry(self, key):
        if key not in self._notifications_queue:
            return

        # Mark it to be cleared
        if self._notifications_queue[key]['last_processed'] is None:
            self._notifications_queue[key]['notify_of_clearance'] = False
        self._notifications_queue[key]['to_clear'] = True
        self._notifications_queue[key]['clear_timestamp'] = time.time()

        notification_entry = self._notifications_queue[key]['instance']
        cool_off_interval = utils.try_parse_int(notification_entry.get_cool_off_interval())
        utils.get_logger().info(f'Notification for {notification_entry.get_entry_type().value} with key \'{key}\' '
                                f'will be cleared after {cool_off_interval} seconds')

    def _clear_notification_entry(self, key):
        if key not in self._notifications_queue:
            return

        notification_entry = self._notifications_queue[key]['instance']
        if self._notifications_queue[key]['notify_of_clearance']:
            self._process_clear_notification(key, notification_entry.get_notification_types(),
                                             notification_entry.get_severity())

        utils.get_logger().info(f'Cleared notification '
                                f'for {notification_entry.get_entry_type().value} with key \'{key}\'')

        # Remove the entry from memory
        self._notifications_queue.pop(key)

    def _main_loop(self):
        while not self._exit_signal.is_set():
            keys_for_entries_to_remove = []
            # for thread safety make a copy of the notification queue before starting with iterating over its contents
            self._mutex.acquire()
            for key in self._notifications_queue:
                notification_entry = self._notifications_queue[key]['instance']
                last_processed = self._notifications_queue[key]['last_processed']
                to_clear = self._notifications_queue[key]['to_clear']
                clear_timestamp = self._notifications_queue[key]['clear_timestamp']
                cool_off_interval = utils.try_parse_int(notification_entry.get_cool_off_interval())

                if to_clear and time.time() - clear_timestamp > cool_off_interval:
                    # Clear the notification
                    keys_for_entries_to_remove.append(key)
                elif not to_clear and last_processed and \
                        notification_entry.get_frequency() == 'interval' and \
                        time.time() - last_processed > notification_entry.get_interval():
                    # This is a recurring notification
                    # Process the notification
                    self._process_notification(key, notification_entry.get_value(),
                                               notification_entry.get_entry_type(),
                                               notification_entry.get_notification_types(),
                                               notification_entry.get_severity(),
                                               notification_entry.get_frequency(),
                                               notification_entry.get_cool_off_interval(),
                                               notification_entry.get_configuration_range(),
                                               notification_entry.get_interval())

                elif not to_clear and last_processed is None and \
                        time.time() - notification_entry.get_timestamp() > cool_off_interval:
                    # This is the first time the notification is processed
                    self._process_notification(key, notification_entry.get_value(),
                                               notification_entry.get_entry_type(),
                                               notification_entry.get_notification_types(),
                                               notification_entry.get_severity(),
                                               notification_entry.get_frequency(),
                                               notification_entry.get_cool_off_interval(),
                                               notification_entry.get_configuration_range(),
                                               notification_entry.get_interval())

            if len(keys_for_entries_to_remove) > 0:
                for key in keys_for_entries_to_remove:
                    self._clear_notification_entry(key)

            self._mutex.release()

            time.sleep(1)  # Sleep for one second

    def finalize(self):
        if not self._options.notifications_module:
            return

        self._exit_signal.set()

        self._mutex.acquire()
        if len(self._notifications_queue) > 0:
            self._notifications_queue.clear()
        self._mutex.release()
